import re

from typing import TYPE_CHECKING, Any, Dict
from urllib.parse import urlparse

from pydantic import AnyUrl

if TYPE_CHECKING:
    from pydantic.config import BaseConfig
    from pydantic.fields import ModelField


class WasbPath(AnyUrl):
    allowed_schemes = {"https", "wasb", "wasbs", "https", "az", "abfs"}

    __slots__ = ()

    @classmethod
    def validate(
        cls, value: Any, field: "ModelField", config: "BaseConfig"
    ) -> "AnyUrl":
        if isinstance(value, Dict):
            _value = value.get("container")
            if not _value:
                raise ValueError("Received a wrong bucket definition: %s", value)
            if "://" not in _value:
                _value = "wasbs://{}".format(_value)
            storage_account = value.get("storage_account") or value.get(
                "storageAccount"
            )
            if storage_account:
                _value = "{}@{}".format(_value, storage_account)
            path = value.get("path")
            if path:
                _value = "{}/{}".format(_value, path)
            value = _value
        cls._validate(value)
        return super(WasbPath, cls).validate(value=value, field=field, config=config)

    def to_param(self):
        return str(self)

    @staticmethod
    def _validate(value):
        try:
            parsed_url = urlparse(value)
        except Exception as e:
            raise ValueError("Received a wrong url definition: %s" % e)
        if parsed_url.scheme in {"wasb", "wasbs"}:
            match = re.match(
                "([^@]+)@([^.]+)\\.blob\\.core\\.windows\\.net", parsed_url.netloc
            )
            if match is None:
                raise ValueError(
                    "wasbs url must be of the form <container>@<account>.blob.core.windows.net"
                )
            container = match.group(1)
            storage_account = match.group(2)
            path = parsed_url.path
            path = path.strip("/")
        elif parsed_url.scheme == "https":
            match = re.match("([^@]+)\\.blob\\.core\\.windows\\.net", parsed_url.netloc)
            if match is None:
                raise ValueError(
                    "wasbs url must be of the form <container>.blob.core.windows.net"
                )
            storage_account = match.group(1)
            path = parsed_url.path
            path = path.strip("/")
            if "/" not in path:
                # this means path is the container_name
                container, path = path, ""
            else:
                container, path = path.split("/", 1)
        else:
            storage_account = None
            container = parsed_url.netloc
            path = parsed_url.path or ""
            path = path.strip("/")
        return dict(
            container=container, storage_account=storage_account, path=path.strip("/")
        )

    @property
    def structured(self):
        return self._validate(self.to_param())

    def get_container_path(self):
        structured = self.structured
        if structured.get("path"):
            return f"{structured['container']}/{structured['path']}"
        return structured["container"]



NAME = "clipped"
VERSION = "0.1.0"
DESC = "Common shortcuts and utils."
URL = "https://github.com/mmourafiq/clipped"
AUTHOR = "Mourad Mourafiq"
EMAIL = "mourad.mourafiq@gmail.com"
LICENSE = "Apache 2.0"



import logging
import os
import signal
import threading

from contextlib import contextmanager
from typing import Generator, Optional

_logger = logging.getLogger("clipped.workers")


def get_pool_workers() -> int:
    return min(32, (os.cpu_count() or 1) + 4)


def get_core_workers(per_core: int, max_workers: Optional[int] = None) -> int:
    count = per_core * (os.cpu_count() or 1) + 1
    return max(count, max_workers) if max_workers else count


@contextmanager
def exit_context() -> Generator:
    exit_event = threading.Event()

    def _exit_handler(*args, **kwargs) -> None:
        _logger.info("Keyboard Interrupt received, exiting pool.")
        exit_event.set()

    original = signal.getsignal(signal.SIGINT)
    try:
        signal.signal(signal.SIGINT, _exit_handler)
        yield exit_event
    except SystemExit:
        pass
    finally:
        signal.signal(signal.SIGINT, original)


def get_wait(current: int) -> float:
    intervals = [0.25, 0.5, 1.0, 2.0, 4.0, 8.0]

    if current >= 5:
        current = 5
    return intervals[current]



from typing import TYPE_CHECKING

from clipped.config.constants import PARAM_REGEX

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator


def validate_image(image, allow_none=False):
    if not image:
        if allow_none:
            return
        else:
            raise ValueError("Image is required")

    if isinstance(image, dict):  # Backwards compatibility
        image = image.get("name")
    if not isinstance(image, str):
        raise TypeError("Image value should be a string")
    param = PARAM_REGEX.search(image)
    if param:
        return image
    if " " in image:
        raise ValueError("Invalid docker image `{}`".format(image))
    _image = image.strip("https://")
    _image = _image.strip("http://")
    tagged_image = _image.split(":")
    if len(tagged_image) > 3:
        raise ValueError("Invalid docker image `{}`".format(image))
    if len(tagged_image) == 3 and (
        "/" not in tagged_image[1] or tagged_image[1].startswith("/")
    ):
        raise ValueError("Invalid docker image `{}`".format(image))

    return image


class ImageStr(str):
    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        return validate_image(value)



from typing import Any, Callable, Optional

import orjson


def orjson_dumps(
    obj: Any,
    *,
    option: Optional[int] = orjson.OPT_SERIALIZE_NUMPY,
    default: Optional[Callable[[Any], Any]] = None,
) -> str:
    return orjson.dumps(obj, default=default, option=option).decode()



import json
import sys
import yaml

from typing import Dict, List, Optional, Union

import click

from rich import box
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.syntax import Syntax
from rich.table import Column, Table
from rich.theme import Theme

from clipped.utils.humanize import humanize_attrs
from clipped.utils.lists import to_list
from clipped.utils.units import to_unit_memory


class Printer:
    COLORS = ["yellow", "blue", "magenta", "green", "cyan", "red", "white"]
    console = Console(
        theme=Theme(
            {
                "header": "yellow",
                "success": "green",
                "info": "cyan",
                "warning": "magenta",
                "error": "red",
                "white": "white",
            }
        ),
        markup=True,
    )

    @staticmethod
    def get_progress():
        return Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
            TaskProgressColumn(),
            TextColumn("eta"),
            TimeRemainingColumn(),
            TextColumn("elapsed"),
            TimeElapsedColumn(),
        )

    @classmethod
    def get_live(cls):
        return Live(console=cls.console)

    @staticmethod
    def get_table(*args, **kwargs):
        return Table(*[Column(header=h, no_wrap=True) for h in args], **kwargs)

    @staticmethod
    def pprint(value):
        """Prints as formatted JSON"""
        click.echo(json.dumps(value, sort_keys=True, indent=4, separators=(",", ": ")))

    @classmethod
    def print_md(cls, md: str):
        cls.console.print(Markdown(md))

    @classmethod
    def print_text(cls, value: str):
        syntax = Syntax(value, "txt", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_yaml(cls, value: any):
        if isinstance(value, str):
            value = yaml.safe_load(value)
        value = yaml.safe_dump(value, sort_keys=True, indent=2)
        syntax = Syntax(value, "yaml", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_json(cls, value: any):
        if isinstance(value, str):
            value = json.loads(value)
        value = json.dumps(value, sort_keys=True, indent=4, separators=(",", ": "))
        syntax = Syntax(value, "json", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print(cls, text: str):
        cls.console.print(text)

    @classmethod
    def help(cls, command_help: Optional[str] = None):
        if command_help:
            cls.console.print(
                "Please run [white]`{} --help`[/white] for more details".format(
                    command_help
                ),
                style="info",
            )

    @classmethod
    def heading(cls, text: str):
        cls.header("\n{}\n".format(text))

    @classmethod
    def header(cls, text: str):
        cls.console.print(text, style="header")

    @classmethod
    def warning(cls, text: str, command_help: Optional[str] = None):
        cls.console.print(text, style="warning")
        if command_help:
            cls.help(command_help=command_help)

    @classmethod
    def success(cls, text: str):
        cls.console.print(text, style="success")

    @classmethod
    def error(
        cls,
        text: str,
        sys_exit: bool = False,
        command_help: Optional[str] = None,
        **kwargs
    ):
        cls.console.print(text, style="error")
        if command_help:
            cls.help(command_help)
        if sys_exit:
            sys.exit(1)

    @classmethod
    def tip(cls, text: str):
        cls.console.print(text, style="white")

    @classmethod
    def info(cls, text: str):
        cls.console.print(text, style="info")

    @staticmethod
    def add_log_color(value, color):
        return click.style("{}".format(value), fg=color)

    @classmethod
    def add_color(cls, value, style):
        return "[{style}]{value}[/{style}]".format(value=value, style=style)

    @classmethod
    def get_colored_status(cls, status):
        if status == "created":
            return cls.add_color(status, "info")
        elif status == "succeeded":
            return cls.add_color(status, style="success")
        elif status in ["failed", "stopped", "upstream_failed"]:
            return cls.add_color(status, style="error")
        elif status == "done":
            return cls.add_color(status, style="white")

        return cls.add_color(status, style="header")

    @classmethod
    def add_status_color(cls, obj_dict, status_key="status"):
        if obj_dict.get(status_key) is None:
            return obj_dict

        obj_dict[status_key] = cls.get_colored_status(obj_dict[status_key])
        return obj_dict

    @classmethod
    def add_memory_unit(cls, obj_dict, keys):
        keys = to_list(keys)
        for key in keys:
            obj_dict[key] = to_unit_memory(obj_dict[key])
        return obj_dict

    @classmethod
    def decorate_format_value(
        cls, text_format: str, values: Union[List[str], str], color: str
    ):
        values = to_list(values)
        values = [cls.add_color(value, color) for value in values]
        click.echo(text_format.format(*values))

    @staticmethod
    def log(value, nl=False):
        click.echo(value, nl=nl)

    @classmethod
    def dict_tabulate(cls, dict_value: Dict, is_list_dict: bool = False):
        if is_list_dict:
            headers = dict_value[0].keys() if dict_value else []
            table = cls.get_table(*headers)
            for d in dict_value:
                table.add_row(*d.values())
            cls.print(table)
        else:
            table = cls.get_table(show_header=False, padding=0, box=box.SIMPLE)
            for k, v in dict_value.items():
                table.add_row(k, humanize_attrs(k, v))
            cls.print(table)



from clipped.config.constants import NO_VALUE_FOUND
from clipped.config.parser import Parser
from clipped.config.spec import ConfigSpec


class ConfigManager:
    _CONFIG_SPEC = ConfigSpec
    _PARSER = Parser

    def __init__(self, **data):
        self._data = data
        self._requested_keys = set()
        self._secret_keys = set()
        self._local_keys = set()

    @classmethod
    def read_configs(cls, config_values):  # pylint:disable=redefined-outer-name
        config = cls._CONFIG_SPEC.read_from(
            config_values
        )  # pylint:disable=redefined-outer-name
        return cls(**config) if config else None

    def keys_startswith(self, term):
        return [k for k in self._data if k.startswith(term)]

    def keys_endswith(self, term):
        return [k for k in self._data if k.endswith(term)]

    def has_key(self, key):
        return key in self._data

    @property
    def data(self):
        return self._data

    @property
    def requested_keys(self):
        return self._requested_keys

    @property
    def secret_keys(self):
        return self._secret_keys

    @property
    def local_keys(self):
        return self._local_keys

    def get_requested_data(
        self, include_secrets=False, include_locals=False, to_str=False
    ):
        data = {}
        for key in self._requested_keys:
            if not include_secrets and key in self._secret_keys:
                continue
            if not include_locals and key in self._local_keys:
                continue
            value = self._data[key]
            data[key] = "{}".format(value) if to_str else value
        return data

    def get(
        self,
        key: str,
        key_type: str,
        is_list=False,
        is_optional=False,
        is_secret=False,
        is_local=False,
        default=None,
        options=None,
    ):
        """
        Get the value corresponding to the key and converts it to `key_type`/`list(key_type)`.

        Args:
            key: the dict key.
            key_type: one of the supported types.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
             `key_type`: value corresponding to the key parsed to key_type.
        """
        kwargs = {}
        return self._get(
            key=key,
            parser_fct=self._PARSER.parse(key_type),
            is_list=is_list,
            is_optional=is_optional,
            is_secret=is_secret,
            is_local=is_local,
            default=default,
            options=options,
            **kwargs,
        )

    def get_value(self, key):
        return self._data.get(key, NO_VALUE_FOUND)

    def _get(
        self,
        key,
        parser_fct,
        is_list,
        is_optional,
        is_secret,
        is_local,
        default,
        options,
        **kwargs,
    ):
        """
        Get key from the dictionary made out of the configs passed.

        Args:
            key: the dict key.

        Returns:
             The corresponding value of the key if found.

        Raises:
            KeyError
        """
        value = self.get_value(key)
        parsed_value = parser_fct(
            key=key,
            value=value,
            is_list=is_list,
            is_optional=is_optional,
            default=default,
            options=options,
            **kwargs,
        )
        self._add_key(key, is_secret=is_secret, is_local=is_local)
        return parsed_value

    def _add_key(self, key, is_secret=False, is_local=False):
        self._requested_keys.add(key)
        if is_secret:
            self._secret_keys.add(key)
        if is_local:
            self._local_keys.add(key)



from urllib.parse import urlparse


def validate_url(url):
    if not url.startswith(("http://", "https://")):
        return False
    parsed = urlparse(url)
    if not parsed.hostname:
        return False
    return True


URL_FORMAT = "{protocol}://{domain}{path}"



import pytz

from datetime import datetime, timedelta
from typing import Optional

from dateutil.tz import tzlocal

try:
    from django.utils.timezone import now as dj_now  # pylint:disable=import-error
except ImportError:
    dj_now = None


def get_timezone(tz: Optional[str] = None):
    if tz:
        return pytz.timezone(tz)
    return tzlocal()


def now(tzinfo=True, no_micor=False):
    """
    Return an aware or naive datetime.datetime, depending on settings.USE_TZ.
    """
    value = None
    if dj_now:
        try:
            value = dj_now()
        except Exception:  # Improper configuration
            pass
    if not value:
        if tzinfo:
            value = datetime.utcnow().replace(tzinfo=pytz.utc)
        else:
            value = datetime.now()
    if no_micor:
        return value.replace(microsecond=0)
    return value


def local_datetime(datetime_value, tz=None):
    return datetime_value.astimezone(get_timezone(tz))


def get_datetime_from_now(days: int, hours: int = 0, minutes: int = 0) -> datetime:
    return now() - timedelta(days=days, hours=hours, minutes=minutes)



import json
import os
import sys
import yaml

from collections.abc import Mapping
from requests import HTTPError
from typing import Any, Dict
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from clipped.config.exceptions import SchemaError
from clipped.utils.dicts import deep_update
from clipped.utils.lists import to_list


class ConfigSpec:
    _SCHEMA_EXCEPTION = SchemaError

    def __init__(
        self, value: Any, config_type: str = None, check_if_exists: bool = True
    ):
        self.value = value
        self.config_type = config_type
        self.check_if_exists = check_if_exists

    @classmethod
    def get_from(cls, value: Any, config_type: str = None) -> "ConfigSpec":
        if isinstance(value, cls):
            return value

        return cls(value=value, config_type=config_type)

    def check_type(self):
        type_check = self.config_type is None and not isinstance(
            self.value, (Mapping, str)
        )
        if type_check:
            raise self._SCHEMA_EXCEPTION(
                "Expects Mapping, string, or list of Mapping/string instances, "
                "received {} instead".format(type(self.value))
            )

    def read(self) -> Dict:
        if isinstance(self.value, Mapping):
            return self.value

        # try a python file
        if isinstance(self.value, str) and (
            self.config_type == ".py" or ".py" in self.value
        ):
            _f_path, _f_module = self.get_python_file_def(self.value)
            if _f_path and _f_module:
                return self.read_from_python(_f_path, _f_module)

        if os.path.isfile(self.value):
            return self.read_from_file(self.value, self.config_type)

        # try reading a stream of yaml or json
        if not self.config_type or self.config_type in (".json", ".yml", ".yaml"):
            try:
                return self.read_from_stream(self.value)
            except (ScannerError, ParserError):
                raise self._SCHEMA_EXCEPTION(
                    "Received an invalid yaml stream: `{}`".format(self.value)
                )

        if self.config_type == "url":
            return self.read_from_url(self.value)

        if self.config_type == "hub":
            public_hub = self.get_public_registry()
            if public_hub:
                return self.read_from_public_hub(self.value, public_hub)
            return self.read_from_custom_hub(self.value)

        raise self._SCHEMA_EXCEPTION(
            "Received an invalid configuration: `{}`".format(self.value)
        )

    @classmethod
    def read_from(cls, config_values: Any, config_type: str = None) -> Dict:
        """
        Reads an ordered list of configuration values and
        deep merge the values in reverse order.
        """
        if not config_values:
            raise cls._SCHEMA_EXCEPTION(
                "Cannot read config_value: `{}`".format(config_values)
            )

        config_values = to_list(config_values, check_none=True)

        config = {}
        for config_value in config_values:
            config_value = cls.get_from(value=config_value, config_type=config_type)
            config_value.check_type()
            config_results = config_value.read()
            if config_results and isinstance(config_results, Mapping):
                config = deep_update(config, config_results)
            elif config_value.check_if_exists:
                raise cls._SCHEMA_EXCEPTION(
                    "Cannot read config_value: `{}`".format(config_value.value)
                )

        return config

    @classmethod
    def read_from_url(cls, url: str) -> Dict:
        from clipped.utils.requests import safe_request

        resp = safe_request(url)
        resp.raise_for_status()
        return cls.read_from_stream(resp.content.decode())

    @classmethod
    def get_public_registry(cls) -> str:
        return ""

    @classmethod
    def read_from_public_hub(cls, hub: str, public_hub: str) -> Dict:
        hub_values = hub.split(":")
        if len(hub_values) > 2:
            raise cls._SCHEMA_EXCEPTION(
                "Received an invalid hub reference: `{}`".format(hub)
            )
        if len(hub_values) == 2:
            hub_name, version = hub_values
        else:
            hub_name, version = hub_values[0], "latest"
        version = version or "latest"
        url = "{}/{}/{}.yaml".format(public_hub, hub_name, version)
        try:
            return cls.read_from_url(url)
        except HTTPError as e:
            if e.response.status_code == 404:
                raise cls._SCHEMA_EXCEPTION(
                    "Config `{}` was not found, "
                    "please check that the name and tag are valid".format(hub)
                )
            raise cls._SCHEMA_EXCEPTION(
                "Config `{}` could not be fetched, "
                "an error was encountered {}".format(hub, e)
            )

    @classmethod
    def read_from_custom_hub(cls, hub: str) -> Dict:
        raise cls._SCHEMA_EXCEPTION(
            "Config `{}` could not be loaded, "
            "please set a valid registry to load the configs from".format(hub)
        )

    @classmethod
    def read_from_stream(cls, stream) -> Dict:
        results = cls.read_from_yml(stream, is_stream=True)
        if not results:
            results = cls.read_from_json(stream, is_stream=True)
        return results

    @classmethod
    def get_python_file_def(cls, f_path):
        if not isinstance(f_path, str) or ".py" not in f_path:
            return None, None
        results = f_path.split(":")

        if len(results) == 1:  # Default case
            return f_path, "main"

        if len(results) != 2 or not results[1]:
            return None, None

        _f_path = results[0].strip("")
        _module_name = results[1].strip("")
        if not os.path.exists(_f_path):
            raise cls._SCHEMA_EXCEPTION(
                "Received non existing python file: `{}`".format(f_path)
            )
        if not _module_name:
            raise cls._SCHEMA_EXCEPTION(
                "Received an invalid python module: `{}`".format(f_path)
            )
        return _f_path, _module_name

    @classmethod
    def import_py_module(cls, f_path, f_module):
        import importlib.util

        spec = importlib.util.spec_from_file_location(f_module, f_path)
        if sys.modules.get(spec.name) and sys.modules[
            spec.name
        ].__file__ == os.path.abspath(spec.origin):
            module = sys.modules[spec.name]
        else:
            module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = module
            spec.loader.exec_module(module)
        return module

    @classmethod
    def read_from_python(cls, f_path, f_module):
        f_path = os.path.abspath(f_path)
        file_directory = os.path.dirname(f_path)
        if file_directory not in sys.path:
            sys.path.append(file_directory)

        module_name = os.path.splitext(os.path.basename(f_path))[0]
        module = cls.import_py_module(f_path, module_name)
        return getattr(module, f_module)

    @classmethod
    def read_from_file(cls, f_path, file_type):
        _, ext = os.path.splitext(f_path)
        if ext in (".yml", ".yaml") or file_type in (None, ".yml", ".yaml"):
            return cls.read_from_yml(f_path)
        elif ext == ".json" or file_type == ".json":
            return cls.read_from_json(f_path)
        elif ext == ".py" or file_type == ".py":
            _f_path, _f_module = cls.get_python_file_def(f_path)
            if _f_path and _f_module:
                return cls.read_from_python(_f_path, _f_module)
        raise cls._SCHEMA_EXCEPTION(
            "Expects a file with extension: `.yml`, `.yaml`, `.py`, or `json`, "
            "received instead `{}`".format(ext)
        )

    @classmethod
    def read_from_yml(cls, f_path, is_stream=False):
        try:
            if is_stream:
                return yaml.safe_load(f_path)
            with open(f_path) as f:
                return yaml.safe_load(f)
        except (ScannerError, ParserError) as e:
            raise cls._SCHEMA_EXCEPTION(
                "Received non valid yaml: `%s`.\n" "Yaml error %s" % (f_path, e)
            ) from e

    @classmethod
    def read_from_json(cls, f_path, is_stream=False):
        if is_stream:
            try:
                return json.loads(f_path)
            except ValueError as e:
                raise cls._SCHEMA_EXCEPTION("Json error: %s" % e) from e
        try:
            with open(f_path) as f:
                return json.loads(f.read())
        except ValueError as e:
            raise cls._SCHEMA_EXCEPTION(
                "Received non valid json: `%s`.\n" "Json error %s" % (f_path, e)
            ) from e
